from neuram.regions import (
    sensory_cortex,
    thalamus,
    amygdala,
    prefrontal_cortex,
    hippocampus,
    neocortex,
    cerebellum,
    basal_ganglia,
)

__all__ = [
    "sensory_cortex",
    "thalamus",
    "amygdala",
    "prefrontal_cortex",
    "hippocampus",
    "neocortex",
    "cerebellum",
    "basal_ganglia",
]
